/**
 * BNSim is an open-source, parallel, stochastic, and multi-scale modeling
 * platform which integrates various simulation algorithms, together with
 * genetic circuits and chemotaxis pathway models in a 3D environment.
 *
 * BNSim is developed by CMU SLD Group, and released under GPLv2 license
 * Please check http://www.ece.cmu.edu/~sld/bnsim for more information
 *
 * BNSim is developed using C++ and pthread under Linux and Mac OS
 *
 * If you use it for your work, please cite
 * Wei, Guopeng, Paul Bogdan, and Radu Marculescu. "Efficient Modeling and
 * Simulation of Bacteria-Based Nanonetworks with BNSim." Selected Areas in
 * Communications, IEEE Journal on 31.12 (2013): 868-878.
 *
 * @author      Guopeng (Daniel) Wei  danielwei@cmu.edu
 * @version     2.5
 * @since       1.0
 */

#include "core.h"

using namespace BNSim;
using namespace std;

double Bacteria::Dr = 0.28;  // rotational diffusion constant rad^2/s

Bacteria::Bacteria(myVector3d Position, int radius, double velocity, int myid) {
	this->velocity= velocity;
	position = Position;
	pre_position = Position;
    this->radius = radius;
    hittingtime = -1;
	direction.pos.x = randomUniform()-0.5;direction.pos.y = randomUniform()-0.5;direction.pos.z = randomUniform()-0.5;
	direction.normalize();
	org_direction = direction;
	oldCubeLoc = cubeLoc = getCubeIndex(position.pos);
    if(Sim::param.quorumSensing)
        qs = new QuorumSensingSystem(this);
    if(Sim::param.chemotaxis)
        cs = new ChemotaxisSystem(this);
    runTime = 1;
    this->myid = myid;
    oldCubeRelativeIndex =currentCubeRelativeIndex = getCubeRelativeIndex(position.pos);
    alive = true;
    CCW_pre = false;
}

void Bacteria::tick() {

    if(Sim::param.quorumSensing)
        qs->tick();
    if(Sim::param.chemotaxis) {
        cs->tick();
        // keep running
        if(cs->isCCW() && CCW_pre) {
            //	rotationalDiffusion();
            Propel();
        }
        
        // start to run
        else if(cs->isCCW() && !CCW_pre) {
            rotationalDiffusion();
            Propel();
            CCW_pre = true;
        }
        // tumble
        else if(!cs->isCCW() && CCW_pre) {
            rotatePerp(direction, tumbleAngle());
            CCW_pre = false;
        }
    }
}

/**
 * Causes the cell to rotate such that Var(theta(dt)) = 4*D*dt.
 */
void Bacteria::rotationalDiffusion() {
	double dTheta = randomUniform()*sqrt(2*Dr*Sim::param.timeStep);
	rotatePerp(direction, dTheta);
}

/**
 * Return a tumble angle in radians distributed according to Fig. 3, 'Chemotaxis
 * in Escherichia Coli', Berg et al. (claim from 'AgentCell: a digital single-cell
 * assay for bacterial chemotaxis', Emonet et al.).
 */
double Bacteria::tumbleAngle() {
	double tumbleShape = 4;
	double tumbleScale = 18.32;
	double tumbleLocation = -4.60;

	double tumbleAngle;
	do {
		tumbleAngle = sampleGamma(tumbleShape, tumbleScale) + tumbleLocation;
	} while (tumbleAngle > 180);

	return (tumbleAngle*3.14/360.0);
}

void Bacteria::tumble() {
	rotatePerp(direction, tumbleAngle());
}

double Bacteria::sampleGamma(double k, double theta) {
	bool accept = false;
	if (k < 1) {
		// Weibull algorithm
		double c = (1 / k);
		double d = ((1 - k) * pow(k, (k / (1 - k))));
		double u, v, z, e, x;
		do {
			u = randomUniform();	v = randomUniform();
			z = -log(u);		e = -log(v);
			x = pow(z, c);
			if ((z + e) >= (d + x)) {
				accept = true;
			}
		} while (!accept);
		return (x * theta);
	} else {
		// Cheng's algorithm
		double b = (k - log(4));
		double c = (k + sqrt(2 * k - 1));
		double lam = sqrt(2 * k - 1);
		double cheng = (1 + log(4.5));
		double u, v, x, y, z, r;
		do {
			u = randomUniform();		v = randomUniform();
			y = ((1 / lam) * log(v / (1 - v)));
			x = (k * exp(y));
			z = (u * v * v);
			r = (b + (c * y) - x);
			if ((r >= ((4.5 * z) - cheng)) ||
					(r >= log(z))) {
				accept = true;
			}
		} while (!accept);
		return (x * theta);
	}
}

/**
 * Rotates the vector v by an angle theta in a random direction perpendicular to v.
 * From BSim, Gorochowski, Thomas E., et al. "BSim: an agent-based tool for modeling
 * bacterial populations in systems and synthetic biology." PloS one 7.8 (2012): e42790.
 */
void Bacteria::rotatePerp(myVector3d& v, double theta) {
	/* Obtain a random direction perpendicular to v */
	myVector3d random(0.5-randomUniform(),0.5-randomUniform(),0.5-randomUniform());
	myVector3d randomPerp;
	randomPerp.cross(&v, &random);
	rotate(v, randomPerp, theta);
	// Ensure the vector is unit
	v.normalize();
}

/**
 * Rotates the vector v towards the specified axis by an angle theta.
 * From BSim, Gorochowski, Thomas E., et al. "BSim: an agent-based tool for modeling
 * bacterial populations in systems and synthetic biology." PloS one 7.8 (2012): e42790.
 */
void Bacteria::rotate(myVector3d& v, myVector3d axis, double theta) {
	/* Generate the rotation matrix for rotating about the axis by an angle theta */
	myMatrix3d  r;
	r.setFromAxisAngle(axis.pos.x, axis.pos.y, axis.pos.z, theta);
	/* Apply the rotation */
	r.transform(v);
}

double Bacteria::randomUniform() {
	return ((double) rand() / (RAND_MAX));
}

void Bacteria::advance() {
    unregisterinCube(Sim::univ->getCube(oldCubeLoc),oldCubeRelativeIndex);
    registerinCube(Sim::univ->getCube(cubeLoc),currentCubeRelativeIndex);
    pre_position = position;
    oldCubeLoc = cubeLoc;
    oldCubeRelativeIndex = currentCubeRelativeIndex;
    
    if(hittingtime==-1 && position.pos.x>900 && position.pos.y>900 && position.pos.z>900) {
        hittingtime = Sim::param.time;
        alive  = false;
        Sim::univ->getCube(cubeLoc)->occupancy[currentCubeRelativeIndex]=0;
    }
}

void Bacteria::recover() {
    currentCubeRelativeIndex = oldCubeRelativeIndex;
    position = pre_position;
    currentCubeRelativeIndex = oldCubeRelativeIndex;
}

void Bacteria::Propel() {
	myVector3d velocity1;
	velocity1.scale(velocity*Sim::param.timeStep, &direction); // 20 micrometers/sec
	position.add(&velocity1);

	// check boundaries
	if(position.pos.x > Sim::param.WorldSize->pos.x)  {  position.pos.x = Sim::param.WorldSize->pos.x -1; direction.pos.x=-randomUniform()*0.5; direction.pos.y =0.5-randomUniform(),direction.pos.z =0.5-randomUniform(); direction.normalize();}
	if(position.pos.x < 0) { position.pos.x = 1; direction.pos.x=randomUniform()*0.5; direction.pos.y =0.5-randomUniform(),direction.pos.z =0.5-randomUniform(); direction.normalize(); }
	if(position.pos.y > Sim::param.WorldSize->pos.y) { position.pos.y = Sim::param.WorldSize->pos.y -1; direction.pos.x=0.5-randomUniform(); direction.pos.y =-randomUniform()*0.5,direction.pos.z =0.5-randomUniform(); direction.normalize();}
	if(position.pos.y < 0) { position.pos.y = 1; direction.pos.x=0.5-randomUniform(); direction.pos.y =randomUniform()*0.5,direction.pos.z =0.5-randomUniform(); direction.normalize();}
	if(position.pos.z > Sim::param.WorldSize->pos.z)	{ position.pos.z = Sim::param.WorldSize->pos.z -1; direction.pos.x=0.5-randomUniform(); direction.pos.y =0.5-randomUniform(),direction.pos.z =-randomUniform()*0.5; direction.normalize();}
	if(position.pos.z < 0) { position.pos.z = 1; direction.pos.x=0.5-randomUniform(); direction.pos.y =0.5-randomUniform(),direction.pos.z =randomUniform()*0.5; direction.normalize();}

	cubeLoc = getCubeIndex(position.pos);
    currentCubeRelativeIndex = getCubeRelativeIndex(position.pos);

    // case 1: same cube, different relative index
    if(cubeLoc == oldCubeLoc && currentCubeRelativeIndex!=oldCubeRelativeIndex) {
        Sim::univ->getCube(oldCubeLoc)->mut.lock();
        EnvCube* newCube = Sim::univ->getCube(cubeLoc);
        if(newCube->occupancy[currentCubeRelativeIndex]) {
            recover();
        }
        else {
            advance();
        }
        Sim::univ->getCube(oldCubeLoc)->mut.unlock();
    }
    // case 2: different cube, different relative index
    else if(cubeLoc!=oldCubeLoc && currentCubeRelativeIndex!=oldCubeRelativeIndex) {
        Sim::univ->getCube(oldCubeLoc)->mut.lock();
        if(!Sim::univ->getCube(cubeLoc)->mut.try_lock()) {
            recover();
            Sim::univ->getCube(oldCubeLoc)->mut.unlock();
            return;
        }
        
        int tempLoc = oldCubeLoc;
        EnvCube* newCube = Sim::univ->getCube(cubeLoc);
        if(newCube->occupancy[currentCubeRelativeIndex]) {
            recover();
        }
        else {
            advance();
        }
        
        Sim::univ->getCube(tempLoc)->mut.unlock();
        Sim::univ->getCube(cubeLoc)->mut.unlock();
    }
    
}

bool Bacteria::intersect() {

	if(Sim::param.bacteriaRadius== 0) return false;

	EnvCube* newCube = Sim::univ->getCube(cubeLoc);
	cubeLoc = getCubeIndex(position.pos);

	// calculate the relative index in the cube
	int newIndex = getCubeRelativeIndex(position.pos);
	// check if it is myself
	if(newIndex == currentCubeRelativeIndex) {
		return false;
	}
    
	if(newCube->occupancy[newIndex])
		return true;
	else
		return false;
}

int Bacteria::getCubeIndex(doubleVector3d pos) {
    intVector3d loc;
    
	double x = pos.x/(double)(Sim::param.cubeSize.x);
	double y = pos.y/(double)(Sim::param.cubeSize.y);
	double z = pos.z/(double)(Sim::param.cubeSize.z);
    
	double remainderx = (int)(x*100) % 100;
	double remaindery = (int)(y*100) % 100;
	double remainderz = (int)(z*100) % 100;
	if(remainderx>=50)		loc.x = floor(x)+1;
	else 	loc.x = floor(x);
	if(remaindery>=50)		loc.y = floor(y)+1;
	else 	loc.y = floor(y);
	if(remainderz>=50)		loc.z = floor(z)+1;
	else 	loc.z = floor(z);
    
	if(loc.x == Sim::param.cubeNumber.x) loc.x  = Sim::param.cubeNumber.x-1;
	if(loc.y == Sim::param.cubeNumber.y) loc.y  = Sim::param.cubeNumber.y-1;
	if(loc.z == Sim::param.cubeNumber.z) loc.z  = Sim::param.cubeNumber.z-1;
    
    int cubeIndex = loc.x*Sim::param.cubeNumber.y*Sim::param.cubeNumber.z + loc.y*Sim::param.cubeNumber.z + loc.z;
    
    return cubeIndex;
}

int Bacteria::getCubeRelativeIndex(doubleVector3d pos) {
    int newIndexX= ((int)(floor(pos.x))%Sim::param.cubeSize.x)/radius;
	int newIndexY= ((int)(floor(pos.y))%Sim::param.cubeSize.y)/radius;
	int newIndexZ= ((int)(floor(pos.z))%Sim::param.cubeSize.z)/radius;
    
    int cubeRelative =newIndexX*Sim::param.cubeSize.y*Sim::param.cubeSize.z/(radius*radius) + newIndexY*Sim::param.cubeSize.z/radius + newIndexZ;
    
    return cubeRelative;
}

void Bacteria::registerinCube(EnvCube* cube, int cubeRelativeIndex) {
    cube->particleNumber++;
    if(Sim::param.bacteriaRadius!= 0) {
    	cube->occupancy[cubeRelativeIndex]=1;
        if(Sim::param.RecordVisit) cube->visited[cubeRelativeIndex]=1;
    }
}

void Bacteria::unregisterinCube(EnvCube* cube, int cubeRelativeIndex) {
     cube->particleNumber--;
	 if(Sim::param.bacteriaRadius!= 0) {
		 cube->occupancy[cubeRelativeIndex]=0;
	 }
}
