
/**
 * BNSim is an open-source, parallel, stochastic, and multi-scale modeling
 * platform which integrates various simulation algorithms, together with
 * genetic circuits and chemotaxis pathway models in a 3D environment.
 *
 * BNSim is developed by CMU SLD Group, and released under GPLv2 license
 * Please check http://www.ece.cmu.edu/~sld/bnsim for more information
 *
 * BNSim is developed using C++ and pthread under Linux and Mac OS
 *
 * If you use it for your work, please cite
 * Wei, Guopeng, Paul Bogdan, and Radu Marculescu. "Efficient Modeling and
 * Simulation of Bacteria-Based Nanonetworks with BNSim." Selected Areas in
 * Communications, IEEE Journal on 31.12 (2013): 868-878.
 *
 * @author      Guopeng (Daniel) Wei  danielwei@cmu.edu
 * @version     2.5
 * @since       1.0
 */

#ifndef CORE_H_
#define CORE_H_

#include <cmath>
#include <string>
#include <random>
#include <cstdlib>
#include <vector>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <fstream>
#include <mutex>

using namespace std;

namespace BNSim {

class myVector3d;
class Bacteria;
class Universe;
class ChemotaxisSystem;
class QuorumSensingSystem;
class Bacteria;

struct chemicalSpecies {
    double diffusionCoefficient;
    double degradationRate;
    string name;
    double totalamount;
};
    
struct intVector3d	{
	int x,y,z;
};

struct doubleVector3d	{
	double x,y,z;
};

struct Param {
	// some constants
	double timeStep; //= 0.01; // s
	double viscosity; //= 2.7e-3;
	double diffusionC; //= 890; // microns^2/
	myVector3d* WorldSize;
	intVector3d cubeNumber, cubeSize;
	unsigned int ThreadNumber;
	unsigned int ChemicalNumber;
	bool leakyBoundary;
	double time;
	bool calibrate;
	string workdir;
	bool environmentOnly;
	bool bacteriaOnly;
    bool networkExport;
    int bacteriaRadius;
    bool RecordVisit;
    double QST;
    double QSD;
    bool quorumSensing;
    bool chemotaxis;
    int bacNum;
	Param(void);
};

class myMatrix3d {
	public:
	myMatrix3d();
	~myMatrix3d();
	void transform(myVector3d& t);
	void setFromAxisAngle(double x, double y, double z, double angle);

	private:
	double m00, m01, m02, m10, m11, m12, m20, m21, m22;
};

class myVector3d {
	public:
	myVector3d();
	myVector3d(double x, double y, double z);
	myVector3d(const myVector3d* v);
	~myVector3d();
	myVector3d& operator=(myVector3d rhs) { pos.x = rhs.pos.x; pos.y = rhs.pos.y; pos.z = rhs.pos.z; return *this;};
	void normalize();
	double length();
	double lengthSquared();
	void add(const myVector3d* v);
	void sub(const myVector3d* v);
	void negate(const myVector3d* v);
	void scale(double s);
	void scale(double s, const myVector3d* v);
	void cross(const myVector3d* v1, const myVector3d* v2);
	doubleVector3d pos;
};


class NodeCoverageExporter {
public:
	NodeCoverageExporter();
	double update();
	void close();
private:
	char file_name[100];
	ofstream nodeC;
};

class exporter {
public:
	static void dump_Gradient(int chemicalIndex);
	static void dump_Bacteria(void);
    static void dump_HittingTime(void);
};

struct thread_data_t {
  unsigned int start;
  unsigned int end;
};


class EnvCube {
public:
	vector<double> chemicals;
	vector<double> preChemicals;
	int particleNumber;
	vector<unsigned char> occupancy;
	vector<unsigned char> visited;
	EnvCube();
    mutex mut;
};

class Sim {
public:
	static struct Param param;
	static Universe* univ;
	static vector<EnvCube *> cubesList;
	static vector<Bacteria *> BacteriaList;
	static vector<chemicalSpecies> chemical;
	void tick(void);
};

class Bacteria {
public:
	Bacteria(myVector3d Position,  int radius, double velocity, int id);
	myVector3d* getPos() {return &position;}
	int getCubeLoc() {return cubeLoc;}
	void setCubeLoc(int loc) {cubeLoc = loc;}
	void setOldCubeLoc(int loc) { oldCubeLoc = loc;}
	void tick();
	bool intersect();
	void registerinCube(EnvCube* cube, int cubeRelativeIndex);
	void unregisterinCube(EnvCube* cube, int cubeRelativeIndex);
	int getID() {return myid;}
    QuorumSensingSystem* getQS() { return qs;}
    int getCubeRelativeIndex(doubleVector3d pos);
    int getCubeIndex(doubleVector3d pos);
    int currentCubeRelativeIndex,oldCubeRelativeIndex, cubeLoc, oldCubeLoc;
    int getCurrentCubeRelativeIndex() {return currentCubeRelativeIndex;}
    int myid;
    bool isalive() {return alive;}
    int getHittingTime() { return hittingtime;}
private:
	myVector3d direction, org_direction, position, pre_position;
	QuorumSensingSystem* qs;
    ChemotaxisSystem* cs;
    bool CCW_pre;
	double velocity;
    bool alive;
	static double Dr;  // rotational diffusion constant rad^2/s
    double runTime;
	void rotationalDiffusion();
	double tumbleAngle();
	void tumble();
	void Propel();
    void recover();
    void advance();
	void rotate(myVector3d& v, myVector3d axis, double theta);
	void rotatePerp(myVector3d& v, double theta) ;
	double sampleGamma(double k, double theta);
	double randomUniform();
    int radius;
    int hittingtime;
};

class Universe {
private:
	pthread_t *thr;
	thread_data_t *thr_data;
	thread_data_t *evn_thr_data;
	void update_environment_p(void);
	void update_bacteria_parallel(void);
public:
	Universe();
	~Universe(void);
	void reset();
//    BacteriaNetwork* BacNetwork;
    EnvCube* getCube(const intVector3d cubeLocation);
	EnvCube* getCube(int x, int y, int z);
	EnvCube* getCube(int index);
	const intVector3d getcubeLocation(const myVector3d* location);
	Bacteria* getBacteria(int index);
	void addChemical(int index, double con);
	void addBacteria(Bacteria *bacteria);
	void addChemical(double c);
	void prepare_multithreading(void);
	void tick(void);
	int checkChemical(string name);
};

class ChemotaxisSystem {
public:
    ChemotaxisSystem(Bacteria * host);
    void tick();
    bool isCCW() {return CCW;}
private:
    double Y;
    double Ya;
    double Ye;
    double H;
    double inverseTau;
    double kr;
    double kb;
    double m;
    double m0;
    double aspcon;
    double activity;
    double alpha;
    double KA;
    double KI;
    int N_tar;
    double Kd;
    double g0;
    double g1;
    double w;
    bool CCW;
    Bacteria * host;
    double randomUniform();
    
    // motor
	double tau;
    double runtime;
    double tumbletime;
    double kPlus();
    double kMinus();
};

class QuorumSensingSystem {
private:

	// define molecules

	// las
	double A1;
	double C1;
	double S;
	double R1;
	double R1A1;
	double R1QSI1;

	// rhl
	double A2;
	double C2;
	double R2;
	double R2A2;

	// QSI
	double qsi1;
	double qsi2;
	double qsi3;

	// define rate constants
	double kA1;
	double kA2;
	double kR1;
	double kR2;
	double kS;
	double CR1;
	double CR2;
	double CA1;
	double CA2;
	double CS;

	// define binding constants
	double KL1;
	double KL2;
	double KL3;
	double KL4;
	double KS1;
	double KS;
	double KR1;
	double k[13];

	Bacteria * host;
    double randomUniform();
public:
	QuorumSensingSystem(Bacteria * host);
	double getA1() { return A1;}
	double getR1() { return R1;}
	double getA2() { return A2;}
	double getR2() { return R2;}
    double getC1() { return C1;}
	double getC2() { return C2;}
	void tick();
	bool resistantMutant;
    bool activated;
};

}

#endif
